/* include the standard headers for this project */
#include "headers.h"

/* Abstract operations on HDF-EOS(5) objects
   (files, swaths, fields)
   primarily attribute inquiry and fetching
*/

/* the types of HDF-EOS5 objects we're going to mung */
/* (see headers.h) */

int attribute_names(hid_t handle, he5_object htype, char const *field,
		       char *namebuf, unsigned long *bufsize)
{
  switch(htype) {
  case he5_file:
    return HE5_EHinqglbattrs(handle, namebuf, (long *)bufsize);
  case he5_swath:
    return HE5_SWinqattrs(handle, namebuf, (long *)bufsize);
  case he5_geofield:
  case he5_datafield:
    return HE5_SWinqlocattrs(handle, field, namebuf, (long *)bufsize);
  default:
    fatal_error("Unknown he5 abstract type", "Ugly Internal Error");
    return 0;
  }
}

herr_t attribute_info(hid_t handle, he5_object htype, char const *field,
		      char const *attname,
		      H5T_class_t *type, hsize_t *count, size_t *size, H5T_sign_t *signp)
{

  switch(htype) {
  case he5_file:
    {
      herr_t result = HE5_EHglbattrinfo(handle, attname, type, count);

      *type = read_in_type(*type);
      *type = type_info(*type, size, signp);

      return result;
    }
  case he5_swath:
    {
      herr_t result = HE5_SWattrinfo(handle, attname, type, count);

      *type =  read_in_type(*type);
      *type = type_info(*type, size, signp);

      return result;
    }
  case he5_geofield:
  case he5_datafield:
    {
      herr_t result =  HE5_SWlocattrinfo(handle, field, attname, type, count);

      *type =  read_in_type(*type);
      *type = type_info(*type, size, signp);
      return result;
    }
  default:
    fatal_error("Unknown he5 abstract type", "Ugly Internal Error");
    return 0;
  }
}

herr_t read_attribute(hid_t handle, he5_object htype, char const *field,
		      char const *attname, void *databuf)
{
  switch(htype) {
  case he5_file:
    return HE5_EHreadglbattr(handle, attname, databuf);
  case he5_swath:
    return HE5_SWreadattr(handle, attname, databuf);
  case he5_geofield:
  case he5_datafield:
    return HE5_SWreadlocattr(handle, field, attname, databuf);
  default:
    fatal_error("Unknown he5 abstract type", "Ugly Internal Error");
    return 0;
  }
}

/* maps platform-specific types to native common types that will be
   actually read in by HDF-EOS reader routines.  This will go away if
   the HDF-EOS developers take my suggestion and have the inquiry
   functions only emit "native" types...
*/

hid_t read_in_type(hid_t type) {
  switch(type) {
    /* the values we're going to map everything to - the "natives" */
  case HE5T_NATIVE_CHAR:  /* This one is the "STRING" value */
  case HE5T_NATIVE_SCHAR:
  case HE5T_NATIVE_UCHAR:
  case HE5T_NATIVE_INT:
  case HE5T_NATIVE_UINT:
  case HE5T_NATIVE_SHORT:
  case HE5T_NATIVE_USHORT:
  case HE5T_NATIVE_LONG:
  case HE5T_NATIVE_ULONG:
  case HE5T_NATIVE_LLONG:
  case HE5T_NATIVE_ULLONG:
  case HE5T_NATIVE_FLOAT:
  case HE5T_NATIVE_DOUBLE:
  case HE5T_NATIVE_LDOUBLE: return type;
    /* signed 8-bit */
  case  HE5T_NATIVE_INT8:
  case  HE5T_STD_I8BE:
  case  HE5T_STD_I8LE: return HE5T_NATIVE_SCHAR;
    /* unsigned 8-bit */
  case  HE5T_NATIVE_UINT8:
  case  HE5T_NATIVE_B8:
  case  HE5T_STD_U8BE:
  case  HE5T_STD_U8LE:
  case  HE5T_STD_B8BE:
  case  HE5T_STD_B8LE: return HE5T_NATIVE_UCHAR;
    /* signed 16-bit */
  case  HE5T_NATIVE_INT16:
  case  HE5T_STD_I16BE:
  case  HE5T_STD_I16LE: return HE5T_NATIVE_SHORT;
    /* unsigned 16-bit */
  case  HE5T_NATIVE_UINT16:
  case  HE5T_NATIVE_B16:
  case  HE5T_STD_U16BE:
  case  HE5T_STD_U16LE:
  case  HE5T_STD_B16BE:
  case  HE5T_STD_B16LE: return HE5T_NATIVE_USHORT;
    /* signed 32-bit */
  case  HE5T_NATIVE_INT32:
  case  HE5T_STD_I32BE:
  case  HE5T_STD_I32LE: return HE5T_NATIVE_INT;
    /* unsigned 32-bit */
  case  HE5T_NATIVE_UINT32:
  case  HE5T_NATIVE_B32:
  case  HE5T_STD_U32BE:
  case  HE5T_STD_U32LE:
  case  HE5T_STD_B32BE:
  case  HE5T_STD_B32LE: return HE5T_NATIVE_UINT;
    /* signed 64-bit */
  case  HE5T_NATIVE_INT64:
  case  HE5T_STD_I64BE:
  case  HE5T_STD_I64LE: return HE5T_NATIVE_LONG;
    /* unsigned 64-bit */
  case  HE5T_NATIVE_UINT64:
  case  HE5T_NATIVE_B64:
  case  HE5T_STD_U64BE:
  case  HE5T_STD_U64LE:
  case  HE5T_STD_B64BE:
  case  HE5T_STD_B64LE: return HE5T_NATIVE_ULONG;
    /* 32-bit float */
  case  HE5T_IEEE_F32BE:              
  case  HE5T_IEEE_F32LE: return HE5T_NATIVE_FLOAT;
    /* 64-bit float */
  case  HE5T_IEEE_F64BE:               
  case  HE5T_IEEE_F64LE: return  HE5T_NATIVE_DOUBLE;
    /* Weird stuff  and default */
  case  HE5T_NATIVE_HSIZE:
  case  HE5T_NATIVE_HERR:
  case  HE5T_NATIVE_HBOOL:
  default: return -1;
  }
}

/* and a little piece of junk that takes the "native" types and breaks out
   important stuff about them
*/
H5T_class_t type_info (hid_t type, unsigned int *size, int *signp)
{
  int hack;
  unsigned int hack2;
  if(size == NULL) size = &hack2;
  if(signp == NULL) signp = &hack;
  switch(type) {
    /* the values we're going to map everything to - the "natives" */
    /* one or both of these should map to H5T_STRING, I think... */
  case HE5T_NATIVE_CHAR:
    *size = 1; *signp = 0; return H5T_STRING;
  case HE5T_NATIVE_SCHAR:
    *size = 1; *signp = 1; return H5T_INTEGER;
  case HE5T_NATIVE_UCHAR:
    *size = 1; *signp = 0; return H5T_INTEGER;
  case HE5T_NATIVE_SHORT:
    *size = 2; *signp = 1; return H5T_INTEGER;
  case HE5T_NATIVE_USHORT:
    *size = 2; *signp = 0; return H5T_INTEGER;
  case HE5T_NATIVE_INT: 
    *size = 4; *signp = 1; return H5T_INTEGER;
  case HE5T_NATIVE_UINT:
    *size = 4; *signp = 0; return H5T_INTEGER;
  case HE5T_NATIVE_LONG:
    *size = 8; *signp = 1; return H5T_INTEGER;
  case HE5T_NATIVE_ULONG:
    *size = 8; *signp = 0; return H5T_INTEGER;
  case HE5T_NATIVE_LLONG:
    *size = 8; *signp = 1; return H5T_INTEGER;
  case HE5T_NATIVE_ULLONG:
    *size = 8; *signp = 0; return H5T_INTEGER;
  case HE5T_NATIVE_FLOAT:
    *size = 4; *signp = 1; return H5T_FLOAT;
  case HE5T_NATIVE_DOUBLE:
    *size = 8; *signp = 1; return H5T_FLOAT;
  case HE5T_NATIVE_LDOUBLE:
    *size = 8; *signp = 1; return H5T_FLOAT;
  default:
    TRACE((stderr, "Bogus value %d to type_info\n", type));
    return -1;
  }
}
