/* include the standard headers for this project */
#include "headers.h"

/* the usual predeclaration of functions */

void grind_saved_values(scew_element const *elt);
void grind_attributes(scew_element const *elt, hid_t handle, he5_object htype, char const *field);
void validate_attribute (hid_t handle, he5_object htype, char const *field, H5T_class_t type, hsize_t count, size_t size, H5T_sign_t signp, scew_element const *att);
long fetch_int_value(scew_element const *elt, H5T_sign_t signp);
void validate_int_values(void *ival, hsize_t count, size_t tsize, H5T_sign_t signp, scew_element const *elt);
double fetch_float_value(scew_element const *elt);
void validate_float_values(void *fval, hsize_t count, size_t tsize, scew_element const *elt);
char const *fetch_string_value(scew_element const *elt);
void validate_string_attribute(char *cval, scew_element const *att);
void grind_swaths(scew_element const *elt, hid_t handle, char *swnames, int numswath);
void validate_swath(hid_t swhandle, char const *swname, scew_element const *elt);
void validate_dimension(scew_element const *dimelt,
			long numdims, char *dimnames, hsize_t *dimsizes,
			int mandatoryp);
void validate_field(hid_t handle, char const *fieldname, he5_object hetype, scew_element const *glfelt);

void report_type_mismatch (char const *wanted, char const *saw);
void report_range_mismatch(char const *type, unsigned int indexp, char const *value, scew_element const *elt);
void report_string_mismatch(char const *value, scew_element const *elt);
void report_size_mismatch (char const *type, unsigned long wanted, size_t saw);

/* finally, the main routine */

/* he5v -x xml-file hdf-eos-file */
/* General philosophy of this program:
   The XML file describes elements in the HDF-EOS file which are examined
   and validated.  If there are extra things in the HDF-EOS file, they
   are not checked, validated, or complained about in any way.
*/
int main(int argc, char **argv)
{
  char *xmlfile = NULL;
  char *hdffile = NULL;
  unsigned int i;
  scew_parser *xmlparser;
  hid_t hdfhandle;
  scew_element const *xmlroot;

  for(i=1; i<(unsigned)argc; i++) {
    char *arg = argv[i];
    if(!strcmp(arg, "-x")) xmlfile = argv[++i];
    else if (arg[0] == '-') {
      fprintf(stderr, "FATAL ERROR: unrecognized option %s\n", arg);
      exit(1);
    }
    else hdffile = arg;
  }

  if(xmlfile == NULL || hdffile == NULL) {
    fprintf(stderr, "USAGE: he5v -x xml-file hdf-eos5-file\n");
    exit(-1);
  }

  xmlparser = scew_parser_create();
  if(!scew_parser_load_file(xmlparser, xmlfile)) {
    XML_Parser Xexp = scew_parser_expat(xmlparser);

    fprintf(stderr, "FATAL ERROR: %s is unreadable or bogus XML: %s, line %d column %d\n",
	    xmlfile, XML_ErrorString(XML_GetErrorCode(Xexp)), XML_GetCurrentLineNumber(Xexp), XML_GetCurrentColumnNumber(Xexp));
    exit(-1);
  }

  SS(hdfhandle = HE5_SWopen(hdffile, H5F_ACC_RDONLY));

  /* Done with argument processing, XML file reading, HDF-EOS5 file opening */
  /* On with the show! */
  
  /* Spit if the top-level element isn't <HDF-EOS-Validation-Definition> */
  xmlroot = scew_tree_root(scew_parser_tree(xmlparser));
  check_element_name(xmlroot, "HDF-EOS-Validation-Definition", 1);

  /* We passed that superficial error check, now to inhale the top-level */
  for(i=0; i< scew_element_count(xmlroot); i++) {
    scew_element const *elt = scew_element_by_index(xmlroot,i);
    char const *elt_name = scew_element_name(elt);

    TRACE((stderr, "Processing element named %s\n", elt_name));

    if(!strcmp(elt_name, "SavedValues")) {
      grind_saved_values(elt);
    }
    else if(!strcmp(elt_name, "Attributes")) {
      push_name("FILE", hdffile);
      grind_attributes(elt, hdfhandle, he5_file, NULL);
      pop_name();
    }

    else if(!strcmp(elt_name, "Swaths")) {
      char *swathnames;
      int numswaths;
      long swathnamesize;

      SS(numswaths = HE5_SWinqswath(hdffile, NULL, &swathnamesize));
      if(scew_element_count(elt) > 0 && numswaths == 0)
	fatal_error("No swaths in file:", hdffile);

      swathnames = newstring((unsigned)swathnamesize + 1);
      SS(numswaths = HE5_SWinqswath(hdffile, swathnames, &swathnamesize));

      grind_swaths(elt, hdfhandle, swathnames, numswaths);

      free(swathnames);
    }
    else {
      fatal_error("Unexpected element inside <HDF-EOS-Validation-Definition>", elt_name);
    }
  }
  return print_final_message();
}

/* <SavedValues>
    <SaveInt name="inttag"/><i>42</i>
    <SaveFloat name="floattag"/><f>2.71828</f>
    <SaveString name="stringtag"/><s>random string</s>
   </SavedValues>
*/
void grind_saved_values(scew_element const *elt)
{
  unsigned int svcount = scew_element_count(elt);
  unsigned int i;

  if((svcount % 2) == 1)
    fatal_error("SavedValues with odd number of name/value entries", "");

  for(i=0; i<svcount; i+=2) {
    scew_element const *svelt = scew_element_by_index(elt, i);
    char const *svelt_name = scew_element_name(svelt);

    if(!strcmp(svelt_name, "SaveInt")) {
      char const *iname = get_attribute_string(svelt, "name");
      long ival = fetch_int_value(scew_element_by_index(elt, i+1), H5T_SGN_2);

      if(iname == NULL)
	fatal_error("SaveInt with no name attribute", "");
      save_int(iname, ival);
    }
    else if(!strcmp(svelt_name, "SaveFloat")) {
      char const *fname = get_attribute_string(svelt, "name");
      double fval = fetch_float_value(scew_element_by_index(elt, i+1));

      if(fname == NULL)
	fatal_error("SaveInt with no name attribute", "");
      save_float(fname, fval);
    }
    else if(!strcmp(svelt_name, "SaveString")) {
      char const *sname = get_attribute_string(svelt, "name");
      char const *sval = fetch_string_value(scew_element_by_index(elt, i+1));

      if(sname == NULL)
	fatal_error("SaveInt with no name attribute", "");
      save_string(sname, sval);
    }
    else
      fatal_error("Element other than SaveInt/SaveFloat/SaveString", svelt_name);
  }
}


void grind_attributes(scew_element const *elt, hid_t handle, he5_object htype, char const *field)
{
  scew_element const *mandatory = scew_element_by_name(elt, "Mandatory");
  scew_element const *optional = scew_element_by_name(elt, "Optional");
  unsigned int i;
  char *attnamebuf;
  int numatts;
  unsigned long bufsize;

  /* fetch the global attribute names */
  SS(attribute_names(handle, htype, field, NULL, &bufsize));
  attnamebuf = newstring(bufsize + 1);
  SS(numatts = attribute_names(handle, htype, field, attnamebuf, &bufsize));

  TRACE((stderr, "Grinding attributes\n"));

  /* mash the mandatory attributes */
  if(mandatory != NULL) {
    for(i=0; i<scew_element_count(mandatory); i++) {
      scew_element const *att = scew_element_by_index(mandatory, i);

      if(!strcmp(scew_element_name(att), "Attribute")) {
	scew_attribute const *nameatt = scew_attribute_by_name(att, "name");

	if(nameatt == NULL)
	  fatal_error("Attribute with no name:", "");
	else {
	  char const *attname = scew_attribute_value(nameatt);
	  H5T_class_t type;
	  hsize_t count;
	  size_t size;
	  H5T_sign_t signp;


	  if(name_in_list(attname, attnamebuf, numatts) < 0) {
	    print_name(stderr);
	    fprintf(stderr, "\n\tMissing mandatory attribute: %s\n",
		    attname);
	    bump_errors();
	  }
	  else {
	    push_name("ATTRIBUTE", attname);
	    SS(attribute_info(handle, htype, field, attname,
			      &type, &count, &size, &signp));
	    validate_attribute(handle, htype, field,
			       type, count, size, signp, att);
	    pop_name();
	  }
	}
      }
      /* should we report non <Attribute ...> elements? */
    }
  }

  /* then the optional attributes */
  if(optional != NULL) {
    for(i=0; i<scew_element_count(optional); i++) {
      scew_element const *att = scew_element_by_index(optional, i);

      if(!strcmp(scew_element_name(att), "Attribute")) {
	scew_attribute const *nameatt = scew_attribute_by_name(att, "name");

	if(nameatt == NULL)
	  fatal_error("Attribute with no name:", "");
	else {
	  char const *attname = scew_attribute_value(nameatt);
	  H5T_class_t type;
	  hsize_t count;
	  size_t size;
	  H5T_sign_t signp;


	  if(name_in_list(attname, attnamebuf, numatts) >= 0) {
	    push_name("OPTIONAL-ATTRIBUTE", attname);
	    SS(attribute_info(handle, htype, field, attname,
			      &type, &count, &size, &signp));
	    validate_attribute(handle, htype, field,
			       type, count, size, signp, att);
	    pop_name();
	  }
	}
      }
      /* should we report non <Attribute ...> elements? */
    }
  }

  /* clean up after ourselves */
  free(attnamebuf);
}

/* Yet another *&^%$#@! function mapping small integers safely to
   strings.  Oh, for a language with REAL enumerated types ...
*/
char const *H5T_class_name(H5T_class_t type);
char const *H5T_class_name(H5T_class_t type)
{
  switch(type) {
  case H5T_STRING:  return "H5T_STRING";
  case H5T_INTEGER: return "H5T_INTEGER";
  case H5T_FLOAT:   return "H5T_FLOAT";
  default:          return "Unknown HDF data type";
  }
}

void validate_attribute (hid_t handle, he5_object htype, char const *field,
			 H5T_class_t ntype, hsize_t count,  size_t size, H5T_sign_t signp, scew_element const *att)
{
  /* the list of value attributes */
  char const *valname[] = {"StringValue", "IntValue", "FloatValue",
			   "SameDataTypeAsField", NULL};
  scew_element const *valspec = NULL;
  int valindex;

  TRACE((stderr, "Validating %s %s\n", current_type(), current_name()));
  for(valindex=0; valname[valindex] != NULL; valindex++)
    if((valspec = scew_element_by_name(att, valname[valindex])) != NULL)
      break;
  if(valspec != NULL) {

    void *valbuf = (void *)malloc((size_t)(count * size + 1));

    /* the '+1' above is to insure that string attributes are properly
       C-style terminated, as is this next statement:
    */
    ((char *)valbuf)[count * size] = 0;

    
    SS(read_attribute(handle, htype, field, current_name(), valbuf));

    TRACE((stderr, "Looking for %s\n", valname[valindex]));
    switch(valindex) {

    case 0: /* StringValue */
      if(ntype == H5T_STRING) {
	validate_string_attribute((char *)valbuf, valspec);
      }
      else report_type_mismatch(valname[valindex], "string");
      break;

    case 1: /* IntValue */
      if(ntype == H5T_INTEGER) {
	validate_int_values(valbuf, count, size, signp, valspec);
      }
      else report_type_mismatch(valname[valindex], "integer");
      break;

    case 2: /* FloatValue */
      if(ntype == H5T_FLOAT) {
	validate_float_values(valbuf, count, size, valspec);
      }
      else report_type_mismatch(valname[valindex], "float");
      break;

    case 3: /* SameDataTypeAsField */
      if (htype == he5_geofield || htype == he5_datafield) {
	int numdims;
	hid_t fieldntype;
	H5T_class_t fieldclass;

	/* grumble, grumble, non-const-using HDF-EOS developers ... */
	SS(HE5_SWfieldinfo(handle, (char *)field, &numdims, NULL, &fieldntype, NULL, NULL));
	fieldclass = type_info(read_in_type(fieldntype), NULL, NULL);

	if(ntype != fieldclass) {
	  char msgbuf[100];
	  sprintf(msgbuf, "SameDataTypeAsField(%s)", H5T_class_name(ntype));
	  report_type_mismatch(msgbuf, H5T_class_name(fieldclass));
	}
      }
      else {
	print_name(stderr);
	fatal_error("\nSameTypeAsDataField on a non-field", "");
      }
      break;
    }
    free(valbuf);
  }
  else {
    print_scew_element(att, stderr);
    fatal_error("\n\tNo Value specifier",
		"(StringValue, IntValue, FloatValue, SameDataTypeAsField)");
  }
}

/* <IntValue>
      <i>42</i>
      <IntRange><MinInt><i>1</i></MinInt><MaxInt><i>12</i></MaxInt></IntRange>
      <SaveInt name="tag"/>
      <GetInt name="tag"/>
   </IntValue>
*/

/* A support function that expects to be handed
   either <i>42</i> or <GetInt name="tag"/>
   It returns the (long) value or spits appropriately
*/

long fetch_int_value(scew_element const *elt, H5T_sign_t signp)
{
  char const *ename = scew_element_name(elt);
  long result = 0;

  if(!strcmp(ename, "i")) {
    int bogus;
    char const *stringint = scew_element_contents(elt);

    if(stringint == NULL)
      fatal_error("<i> with no content", "");

    result = careful_atol(stringint, signp, &bogus);
    if(bogus)
      fatal_error("Not a valid integer", stringint);
  }
  else if(!strcmp(ename, "GetInt")) {
    char const *tag = get_attribute_string(elt, "name");
    long const *resptr;

    if(tag == NULL)
      fatal_error("GetInt with no name attribute", "");
    
    resptr = get_saved_int(tag);

    if(resptr == NULL)
      fatal_error("No saved Int value for tag", tag);

    result = *resptr;
    TRACE((stderr, "GetInt of %s is %ld\n", tag, result));
  }
  else
    fatal_error("Expected <i> or <GetInt>, saw instead", ename);

  return result;
}

void validate_int_values(void *ival, hsize_t count, size_t tsize, H5T_sign_t signp, scew_element const *elt)
{
  struct intpair {
    long min;
    long max;
  };
  struct intpair ipbuf[20];
  struct intpair *ipvec;
  unsigned int ipcount = scew_element_count(elt);
  unsigned int i;
  char valstring[25];
  char const * sizestring = get_attribute_string(elt, "size");
  char const *tagname = NULL;

  if(sizestring != NULL) {
    int bogus;
    unsigned long xsize = (unsigned long)careful_atol(sizestring, H5T_SGN_NONE, &bogus);
    if(!bogus) {
      if(xsize != tsize)
	report_size_mismatch("integer", xsize, tsize);
    }
    else fatal_error("Not an integer size", sizestring);
  }

  if(ipcount > (sizeof(ipbuf)/sizeof(struct intpair)))
    ipvec = calloc(ipcount, sizeof(struct intpair));
  else
    ipvec = ipbuf;

  for(i=0; i<ipcount; i++) {
    scew_element const *subelt = scew_element_by_index(elt, i);
    char const *subelt_name = scew_element_name(subelt);

    if(!strcmp(subelt_name, "IntRange")) {
      unsigned int j;

      ipvec[i].min = (signp ? LONG_MIN : 0);
      ipvec[i].max = (signp ? LONG_MAX : ULONG_MAX);

      for(j=0; j<scew_element_count(subelt); j++) {
	scew_element const *mmelt = scew_element_by_index(subelt, j);
	char const *mmname = scew_element_name(mmelt);

	/* this element had better contain exactly one subelement */
	if(scew_element_count(mmelt) != 1)
	  fatal_error("Element must contain exactly one integer", mmname);

	if(!strcmp(mmname, "MinInt")) {
	  ipvec[i].min = fetch_int_value(scew_element_by_index(mmelt, 0), signp);
	}
	else if(!strcmp(mmname, "MaxInt")) {
	  ipvec[i].max = fetch_int_value(scew_element_by_index(mmelt, 0), signp);
	}
	else
	  fatal_error("Unexpected element in IntRange", mmname);
      }
    }
    else if (!strcmp(subelt_name, "SaveInt")) {
      /* just remember this tag name for later */
      /* We can save the value when we know what it is, later on ... */
      tagname = get_attribute_string(subelt, "name");
      /* Spit if there isn't a name */
      if(tagname == NULL)
	fatal_error("No name attribute in SaveInt", "");
      /* and create an ipvec entry that doesn't care */
      ipvec[i].min = (signp ? LONG_MIN : 0);
      ipvec[i].max = (signp ? LONG_MAX : ULONG_MAX);
    }
    else
      ipvec[i].min = ipvec[i].max = fetch_int_value(subelt, signp);
  }

  if(ipcount) {
    if (signp) {
      for(i=0; i<count; i++) {
	long val = 0;
	unsigned int j;
	int valueok = 0;

	switch(tsize) {
	case 1: val = ((char *)ival)[i]; break;
	case 2: val = ((short *)ival)[i]; break;
	case 4: val = ((int *)ival)[i]; break;
	case 8: val = ((long *)ival)[i]; break;
	default:
	  sprintf(valstring, "%d", tsize);
	  fatal_error("Unknown int size:", valstring);
	}
	/* Handle SaveInt here, while we've got what we need */
	if(i == 0 && tagname != NULL)
	  save_int(tagname, val);

	for(j=0; j<ipcount; j++) {
	  if(val >= ipvec[j].min && val <= ipvec[j].max) {
	    valueok = 1;
	    break;
	  }
	}

	if(!valueok) {
	  sprintf(valstring, "%ld", val);
	  report_range_mismatch("int", i, valstring, elt);
	}
      }
    }
    else {
      for(i=0; i<count; i++) {
	unsigned long val = 0;
	unsigned int j;
	int valueok = 0;

	switch(tsize) {
	case 1: val = ((unsigned char *)ival)[i]; break;
	case 2: val = ((unsigned short *)ival)[i]; break;
	case 4: val = ((unsigned int *)ival)[i]; break;
	case 8: val = ((unsigned long *)ival)[i]; break;
	default:
	  sprintf(valstring, "%d", tsize);
	  fatal_error("Unknown int size:", valstring);
	}
	for(j=0; j<ipcount; j++) {
	  if(val >= (unsigned long)ipvec[j].min
	     && val <= (unsigned long)ipvec[j].max) {
	    valueok = 1;
	    break;
	  }
	}

	if(!valueok) {
	  sprintf(valstring, "%lud", val);
	  report_range_mismatch("unsigned int", i, valstring, elt);
	}
      }
    }
    if(ipvec != ipbuf) free(ipvec);
  }
}
/* <FloatValue>
      <f>42</f>
      <FloatRange>
        <MinFloat><i>1</i></MinFloat>
	<MaxFloat><i>12</i></MaxFloat>
      </FloatRange>
      <SaveFloat name="tag"/>
      <GetFloat name="tag"/>
   </FloatValue>
*/

/* A support function that expects to be handed
   either <f>42</f> or <GetFloat name="tag"/>
   It returns the (double) value or spits appropriately
*/

double fetch_float_value(scew_element const *elt)
{
  char const *ename = scew_element_name(elt);
  double result = 0;

  if(!strcmp(ename, "f")) {
    char *bogus;
    char const *stringfloat = scew_element_contents(elt);

    if(stringfloat == NULL)
      fatal_error("<f> with no content", "");

    result = strtod(stringfloat, &bogus);
    if(bogus == stringfloat)
      fatal_error("Not a valid floating-point number", stringfloat);
  }
  else if(!strcmp(ename, "GetFloat")) {
    char const *tag = get_attribute_string(elt, "name");
    double const *resptr;

    if(tag == NULL)
      fatal_error("GetFloat with no name attribute", "");
    
    resptr = get_saved_float(tag);

    if(resptr == NULL)
      fatal_error("No saved Float value for tag", tag);

    result = *resptr;
    TRACE((stderr, "GetFloat of %s is %f\n", tag, result));
  }
  else
    fatal_error("Expected <f> or <GetFloat>, saw instead", ename);

  return result;
}

void validate_float_values(void *fval, hsize_t count, size_t tsize, scew_element const *elt)
{
  struct floatpair {
    double min;
    double max;
  };
  struct floatpair fpbuf[20];
  struct floatpair *fpvec;
  unsigned int fpcount = scew_element_count(elt);
  char valstring[25];
  char const *typename = (tsize == 4 ? "float" : "double");
  unsigned int i;
  char const * sizestring = get_attribute_string(elt, "size");
  char const *tagname = NULL;

  if(sizestring != NULL) {
    int bogus;
    unsigned long xsize = (unsigned long)careful_atol(sizestring, H5T_SGN_NONE, &bogus);
    if(!bogus) {
      if(xsize != tsize)
	report_size_mismatch("float", xsize, tsize);
    }
    else fatal_error("Not a float size", sizestring);
  }

  if(fpcount > (sizeof(fpbuf)/sizeof(struct floatpair)))
    fpvec = calloc(fpcount, sizeof(struct floatpair));
  else
    fpvec = fpbuf;

  for(i=0; i<fpcount; i++) {
    scew_element const *subelt = scew_element_by_index(elt, i);
    char const *subelt_name = scew_element_name(subelt);

    if(!strcmp(subelt_name, "FloatRange")) {
      unsigned int j;

      fpvec[i].min = -HUGE_VAL;
      fpvec[i].max = HUGE_VAL;

      for(j=0; j<scew_element_count(subelt); j++) {
	scew_element const *mmelt = scew_element_by_index(subelt, j);
	char const *mmname = scew_element_name(mmelt);

	/* this element had better contain exactly one subelement */
	if(scew_element_count(mmelt) != 1)
	  fatal_error("Element must contain exactly one float", mmname);

	if(!strcmp(mmname, "MinFloat")) {
	  fpvec[i].min = fetch_float_value(scew_element_by_index(mmelt, 0));
	}
	else if(!strcmp(mmname, "MaxFloat")) {
	  fpvec[i].max = fetch_float_value(scew_element_by_index(mmelt, 0));
	}
	else
	  fatal_error("Unexpected element in FloatRange", mmname);
      }
    }
    else if (!strcmp(subelt_name, "SaveFloat")) {
      /* just remember this tag name for later */
      tagname = get_attribute_string(subelt, "name");
      /* Spit if there isn't a name */
      if(tagname == NULL)
	fatal_error("No name attribute in SaveInt", "");
      /* and create an fpvec entry that doesn't care */
      fpvec[i].min = -HUGE_VAL;
      fpvec[i].max =  HUGE_VAL;
    }
    else
      fpvec[i].min = fpvec[i].max = fetch_float_value(subelt);
  }

  if(fpcount) {
    for(i=0; i<count; i++) {
      unsigned int j;
      double val = 0.0;
      int valueok = 0;

      switch(tsize) {
      case 4: val = ((float *)fval)[i]; break;
      case 8: val = ((double *)fval)[i]; break;
      default:
	sprintf(valstring, "%d", tsize);
	fatal_error("Unknown float size:", valstring);
      }
      /* Handle SaveFloat here, while we've got what we need */
      if(i == 0 && tagname != NULL)
	save_float(tagname, val);

      for(j=0; j<fpcount; j++) {
	if(val >= fpvec[j].min && val <= fpvec[j].max) {
	  valueok = 1;
	  break;
	}
      }
      if(!valueok) {
	sprintf(valstring, "%f", val);
	report_range_mismatch(typename, i, valstring, elt);
      }
    }
    if(fpvec != fpbuf) free(fpvec);
  }
}

/* <StringValue>
     <s>first</s>
     <s>second</s>
     <GetString name="tag"/>
     <SaveString name="tag"/>
     <Matches><s>regular-expression</s></Matches>
   </StringValue>
   Regular expressions follow POSIX (see man regcomp(3) on any Linux box)
*/

/* A support function that expects to be handed
   either <s>42</s> or <GetString name="tag"/>
   It returns the (double) value or spits appropriately
*/

char const *fetch_string_value(scew_element const *elt)
{
  char const *ename = scew_element_name(elt);
  char const *result = NULL;

  if(!strcmp(ename, "s")) {
    result = scew_element_contents(elt);

    if(result == NULL)
      fatal_error("<s> with no content", "");
  }
  else if(!strcmp(ename, "GetString")) {
    char const *tag = get_attribute_string(elt, "name");

    if(tag == NULL)
      fatal_error("GetString with no name attribute", "");
    
    result = get_saved_string(tag);

    if(result == NULL)
      fatal_error("No saved String value for tag", tag);

    TRACE((stderr, "GetString of %s is |%s|\n", tag, result));
  }
  else
    fatal_error("Expected <s> or <GetString>, saw instead", ename);

  return result;
}

void validate_string_attribute(char *cval, scew_element const *svelt)
{
  unsigned int i;
  unsigned int savecount = 0;

  TRACE((stderr, "Validating string: \"%s\" against ", cval));
  TRACEWRAP((print_scew_element(svelt, stderr)));
  TRACE((stderr, "\n"));

  for(i = 0; i < scew_element_count(svelt); i++) {
    scew_element const *subelt = scew_element_by_index(svelt, i);
    char const *subelt_name = scew_element_name(subelt);

    if(!strcmp(subelt_name, "Matches")) {
      char const *matchval;

      if(scew_element_count(subelt) != 1)
	fatal_error("Matches must have exactly one element", "");
      
      matchval = fetch_string_value(scew_element_by_index(subelt, 0));

      if(regex_matches(cval, matchval)) return;
    }
    else if(!strcmp(subelt_name, "SaveString")) {
      char const *tagname = get_attribute_string(subelt, "name");

      if(tagname == NULL)
	fatal_error("No name attribute in SaveString", "");

      save_string(tagname, cval);
      savecount++;
    }
    else {
      char const *testval = fetch_string_value(subelt);

      if(!strcmp(cval, testval)) return;
    }
  }
  /* HMMM... if we get here, did we win or lose? */
  if(savecount != i)
    report_string_mismatch(cval, svelt);
}

void grind_swaths(scew_element const *elt, hid_t filehandle, char *swnames, int numswaths)
{
  unsigned int i;

  TRACE((stderr, "In grind_swaths\n"));

  for(i=0; i<scew_element_count(elt); i++) {
    scew_element const *subelt = scew_element_by_index(elt, i);
    char const *subeltname = scew_element_name(subelt);

    if(!strcmp(subeltname, "Every-Swath")) {
      char *swathname = swnames;
      int s;

      for(s=0; s<numswaths; s++) {
	hid_t swhandle;

	end_name(swathname);
	SS(swhandle = HE5_SWattach(filehandle, swathname));
	validate_swath(swhandle, swathname, subelt);
	HE5_SWdetach(swhandle);
	swathname = next_name(swathname);
      }
    }
    else if(!strcmp(subeltname, "Swath-Named")) {
      scew_attribute const *nameatt = scew_attribute_by_name(subelt, "name");

      if(nameatt == NULL)
	fatal_error("Named swath with no name:", "");
      else {
	char const *swname = scew_attribute_value(nameatt);

	if(name_in_list(swname, swnames, numswaths) < 0) {
	  print_name(stderr);
	  fprintf(stderr, "\n\tMissing mandatory swath: %s\n", swname);
	  bump_errors();
	}
	else {
	  hid_t swhandle;
	  SS(swhandle = HE5_SWattach(filehandle, swname));
	  validate_swath(swhandle, swname, subelt);
	  HE5_SWdetach(swhandle);
	}
      }
    }
    else
      fatal_error("Unrecognized Swath specifier:", subeltname);
  }	  
}

void validate_swath(hid_t swhandle, char const *swname, scew_element const *elt)
{
  unsigned int i;

  TRACE((stderr, "Validating swath %s\n", swname));
  push_name("SWATH", swname);
  for(i=0; i < scew_element_count(elt); i++) {    
    scew_element const *subelt = scew_element_by_index(elt,i);
    char const *subelt_name = scew_element_name(subelt);

    if(!strcmp(subelt_name, "Attributes"))
      grind_attributes(subelt, swhandle, he5_swath, NULL);

    else if(!strcmp(subelt_name, "Swath-Dimensions")) {
      long numdims;
      long bufsize;
      hsize_t *dimsizes;
      char *dimnames;
      unsigned int j;
      

      /* Read in the dimension names and sizes */
      SS(numdims = HE5_SWnentries(swhandle, HE5_HDFE_NENTDIM, &bufsize));
      dimnames = newstring((unsigned)(bufsize + 1));
      dimsizes = (hsize_t *) malloc(sizeof(hsize_t) * numdims);
      SS(numdims = HE5_SWinqdims(swhandle, dimnames, dimsizes));

      /* and validate them against the Dimensions entries */
      for(j=0; j< scew_element_count(subelt); j++) {
	scew_element const *manop = scew_element_by_index(subelt, j);
	char const *manopname = scew_element_name(manop);

	if(!strcmp(manopname, "Mandatory-Dimensions")) {
	  unsigned int k;

	  for(k=0; k< scew_element_count(manop); k++) {
	    validate_dimension(scew_element_by_index(manop, k),
			       numdims, dimnames, dimsizes,
			       1);
	  }
	}
	else if(!strcmp(manopname, "Optional-Dimensions")) {
	  unsigned int k;

	  for(k=0; k< scew_element_count(manop); k++) {
	    validate_dimension(scew_element_by_index(manop, k),
			       numdims, dimnames, dimsizes,
			       0);
	  }
	}
	else
	  fatal_error("Not Mandatory or Optional element:", manopname);
      }
      free(dimsizes);
      free(dimnames);
    }
    else if(!strcmp(subelt_name, "GeolocationFields")) {
      /* I do this instead of
         char *glfnames = "";
       to allow glfnames to be char * (instead of const char *).
       It really doesn't matter here, since the null string just gets
       discarded anyway...  *&^%$#@! C compiler anyway.
      */
      char *glfnames = {0};
      int numglf;
      long glfnamesize;
      unsigned int j;

      SS(numglf = HE5_SWnentries(swhandle, HE5_HDFE_NENTGFLD, &glfnamesize));
      if(numglf > 0) {
	glfnames = newstring((unsigned)(glfnamesize + 1));
	SS(numglf = HE5_SWinqgeofields(swhandle, glfnames, NULL, NULL));
      }

      for(j=0; j<scew_element_count(subelt); j++) {
	scew_element const *glelt = scew_element_by_index(subelt, j);
	char const *gleltname = scew_element_name(glelt);

	if(!strcmp(gleltname, "Every-GeolocationField")) {
	  char *glfname = glfnames;
	  int k;

	  for(k=0; k<numglf; k++) {
	    end_name(glfname);
	    push_name("GEOLOCATION_FIELD", glfname);
	    validate_field(swhandle, glfname, he5_geofield, glelt);
	    pop_name();
	    glfname = next_name(glfname);
	  }
	}
	else if (!strcmp(gleltname, "GeolocationField-Named")) {
	  char const *glfname = get_attribute_string(glelt, "name");

	  if(glfname != NULL) {
	    if(name_in_list(glfname, glfnames, numglf) >=0) {
	      push_name("GEOLOCATION_FIELD", glfname);
	      validate_field(swhandle, glfname, he5_geofield, glelt);
	      pop_name();
	    }
	    else {
	      print_name(stderr);
	      fprintf(stderr, "\n\tMissing mandatory geolocation field: %s\n",
		      glfname);
	      bump_errors();
	    }
	  }
	  else
	    fatal_error("Named element with no name", "");
	}
	else
	  fatal_error("Unrecognized GeolocationField element", gleltname);
      }
    }
    else if(!strcmp(subelt_name, "DataFields")) {
      /* I do this instead of
         char *dfnames = "";
       to allow dfnames to be char * (instead of const char *).
       It really doesn't matter here, since the null string just gets
       discarded anyway...  *&^%$#@! C compiler anyway.
      */
      char *dfnames= {0};
      int numdf;
      long dfnamesize;
      unsigned int j;

      SS(numdf = HE5_SWnentries(swhandle, HE5_HDFE_NENTDFLD, &dfnamesize));
      if(numdf > 0) {
	dfnames = newstring((unsigned)(dfnamesize + 1));
	SS(numdf = HE5_SWinqdatafields(swhandle, dfnames, NULL, NULL));
      }

      for(j=0; j<scew_element_count(subelt); j++) {
	scew_element const *dfelt = scew_element_by_index(subelt, j);
	char const *dfeltname = scew_element_name(dfelt);

	if(!strcmp(dfeltname, "Every-DataField")) {
	  char *dfname = dfnames;
	  int k;

	  for(k=0; k<numdf; k++) {
	    end_name(dfname);
	    push_name("DATA_FIELD", dfname);
	    validate_field(swhandle, dfname, he5_datafield, dfelt);
	    pop_name();
	    dfname = next_name(dfname);
	  }
	}
	else if (!strcmp(dfeltname, "DataField-Named")) {
	  char const *dfname = get_attribute_string(dfelt, "name");

	  if(dfname != NULL) {
	    if(name_in_list(dfname, dfnames, numdf) >=0) {
	      push_name("DATA_FIELD", dfname);
	      validate_field(swhandle, dfname, he5_datafield, dfelt);
	      pop_name();
	    }
	    else {
	      print_name(stderr);
	      fprintf(stderr, "\n\tMissing mandatory data field: %s\n",
		      dfname);
	      bump_errors();
	    }
	  }
	  else
	    fatal_error("Named element with no name", "");
	}
	else
	  fatal_error("Unrecognized DataField element", dfeltname);
      }
    }
    else
      fatal_error("Unrecognized Swath element", subelt_name);
  }
  pop_name();
}

void validate_dimension(scew_element const *dimelt,
			long numdims, char *dimnames, hsize_t *dimsizes,
			int mandatoryp)
{
  char const *dimeltname = scew_element_name(dimelt);
	    
  if(!strcmp(dimeltname, "Dimension")) {
    char const *dimname = get_attribute_string(dimelt, "name");

    if(dimname == NULL) {
      print_scew_element(dimelt, stderr);
      fatal_error("Dimension element with no name", "");
    }
    else {
      int dimindex = name_in_list(dimname, dimnames, numdims);

      if(dimindex >= 0) {
	char const *sizestr = get_attribute_string(dimelt, "size");
	char const *minstr = get_attribute_string(dimelt, "min");
	char const *maxstr = get_attribute_string(dimelt, "max");
	
	if (sizestr != NULL) {
	  hsize_t size = atoi(sizestr);
	  
	  if(dimsizes[dimindex] != size) {
	    print_name(stderr);
	    fprintf(stderr, "\n\tDimension %s size %ld is not == %ld\n",
		    dimname, (long)dimsizes[dimindex], (long)size);
	    bump_errors();
	  }
	}
	if(minstr != NULL) {
	  hsize_t min = atoi(minstr);
	  
	  if(dimsizes[dimindex] < min) {
	    print_name(stderr);
	    fprintf(stderr, "\n\tDimension %s size %ld is < %ld\n",
		    dimname, (long)dimsizes[dimindex], (long)min);
	    bump_errors();
	  }
	}
	if(maxstr != NULL) {
	  hsize_t max = atoi(maxstr);
	  
	  if(dimsizes[dimindex] > max) {
	    print_name(stderr);
	    fprintf(stderr, "\n\tDimension %s size %ld is > %ld\n",
		    dimname, (long)dimsizes[dimindex], (long)max);
	    bump_errors();
	  }
	}
      }
      else if(mandatoryp) {
	print_name(stderr);
	fprintf(stderr, "\n\tMissing mandatory dimension: %s\n", dimname);
	bump_errors();
      }
    }
  }
  else
    fatal_error("Not a Dimension spec element", dimeltname);
	
}

void fetch_fieldinfo(hid_t swhandle, char const *fieldname,
		     hid_t *ntype, char **dimlist, hsize_t **dims, int *numdims);
/* C needs nested functions so bad...*/
void fetch_fieldinfo(hid_t swhandle, char const *fieldname,
		     hid_t *ntype, char **dimlist, hsize_t **dims, int *numdims)
{
  long dimlistsize;

  SS(HE5_SWnentries(swhandle, HE5_HDFE_NENTDIM, &dimlistsize));
  *dimlist = newstring((unsigned)(dimlistsize + 1));

  /* It's not *MY* fault that the HDF-EOS developers don't use const
     properly... heck, I don't use it unless forced by another more
     persnickety library like SCEW...
  */
  SS(HE5_SWfieldinfo(swhandle, (char *)fieldname, numdims, NULL, ntype, *dimlist, NULL));
  *dims = malloc(*numdims * sizeof(hsize_t));
  SS(HE5_SWfieldinfo(swhandle, (char *)fieldname, numdims, *dims, ntype, *dimlist, NULL));
}

void *load_field_data (hid_t swhandle, char const *fieldname, size_t tsize, hsize_t *dimsize, int numdims, hsize_t *valcount);
/* junk function to avoid a little repetitious code loading a field */
void *load_field_data (hid_t swhandle, char const *fieldname, size_t tsize, hsize_t *dimsize, int numdims, hsize_t *valcount)
{
  int i;
  void *databuf;

  *valcount = 1;
  for(i=0; i < numdims; i++)
    *valcount *= dimsize[i];
  databuf = (void *)malloc((size_t)(*valcount * tsize));
  if(databuf == NULL)
    fatal_error("Couldn't allocate memory to read field", fieldname);
  SS(HE5_SWreadfield(swhandle, (char *)fieldname, NULL, NULL, NULL, databuf));
  return databuf;
}

void validate_field(hid_t swhandle, char const *fieldname, he5_object hetype, scew_element const *glfelt)
{
  unsigned int i;
  /* These get filled in on the first occurrence of {Dimensions,DataType},
     and the value gets checked later
  */
  char *dimlist = NULL;
  hid_t ntype;
  int numdims;
  hsize_t *dimsize;

  TRACE((stderr, "Validating %s %s\n", current_type(), current_name()));
  for(i=0; i<scew_element_count(glfelt); i++) {    
    scew_element const *subelt = scew_element_by_index(glfelt,i);
    char const *subelt_name = scew_element_name(subelt);

    if(!strcmp(subelt_name, "Attributes"))
      grind_attributes(subelt, swhandle, hetype, fieldname);

    else if(!strcmp(subelt_name, "Dimensions")) {
      unsigned int j;

      if(dimlist == NULL)
	fetch_fieldinfo(swhandle, fieldname, &ntype, &dimlist, &dimsize, &numdims);
      for(j=0; j<scew_element_count(subelt); j++) {
	char const *dname = scew_element_contents(scew_element_by_index(subelt,j));
	if(name_in_list(dname, dimlist, numdims) < 0) {
	  print_name(stderr);
	  fprintf(stderr, "\n\tMissing mandatory dimension %s\n", dname);
	  bump_errors();
	}
      }
    }
    else if(!strcmp(subelt_name, "DataType")) {
      /* There had better be exactly one subelement, which must be
	 one of IntValue/FloatValue/StringValue */
      unsigned int dteltcount = scew_element_count(subelt);

      switch(dteltcount) {
      case 1:
	{
	  scew_element const *typelt = scew_element_by_index(subelt, 0);
	  char const *typeltname = scew_element_name(typelt);
	  char const *realtype;
	  H5T_class_t h5type;
	  size_t tsize;
	  H5T_sign_t signp;

	  if(dimlist == NULL)
	    fetch_fieldinfo(swhandle, fieldname, &ntype, &dimlist, &dimsize, &numdims);

	  switch(h5type = type_info(read_in_type(ntype), &tsize, &signp)) {
	  case H5T_STRING:  realtype = "string"; break;
	  case H5T_INTEGER: realtype = "integer"; break;
	  case H5T_FLOAT:   realtype = "float"; break;
	  default:
	    {
	      char msgbuf[25];
	      sprintf(msgbuf, "unknown HDF type %d", ntype);
	      report_type_mismatch(typeltname, msgbuf);
	    }
	  /* note - non-local exit to go around the loop again */
	  continue;
	  }

	  /* if there are no subelements of the typelt, we don't need to read
	     in the data and can just check the size here
	  */
	  if(scew_element_count(typelt) == 0) {
	    char const *sizestring = get_attribute_string(typelt, "size");
	    if(sizestring != NULL) {
	      int bogus;
	      unsigned long wsize = (unsigned long)careful_atol(sizestring, H5T_SGN_NONE, &bogus);
	      if(!bogus) {
		if(wsize != tsize)
		  report_size_mismatch(typeltname, wsize, tsize);
	      }
	      else
		fatal_error("Not a DataType size", sizestring);
	    }
	  }
	  else {

	    if(!strcmp(typeltname, "IntValue")) {
	      if(h5type != H5T_INTEGER)
		report_type_mismatch(typeltname, realtype);
	      else {
		hsize_t intcount;
		void *databuf = load_field_data(swhandle, fieldname, tsize, dimsize, numdims, &intcount);
		validate_int_values(databuf, intcount, tsize, signp, typelt);
		free(databuf);
	      }
	    }
	    else if(!strcmp(typeltname, "FloatValue")) {
	      if(h5type != H5T_FLOAT)
		report_type_mismatch(typeltname, realtype);
	      else {
		hsize_t intcount;
		void *databuf = load_field_data(swhandle, fieldname, tsize, dimsize, numdims, &intcount);
		validate_float_values(databuf, intcount, tsize, typelt);
		free(databuf);
	      }
	    }
	    else if(!strcmp(typeltname, "StringValue")) {
	      if(h5type != H5T_STRING)
		report_type_mismatch(typeltname, realtype);
	      /* punting StringValue validation - I don't think this can
		 actually happen, anyway ... */
	    }
	    else
	      fatal_error("not a DataType specifier", typeltname);
	  }
	}
	break;
      case 0:
	fatal_error("DataType with no type specifier", "");
	break;
      default:
	print_name(stderr);
	fprintf(stderr, "\n\tFATAL ERROR: More than one DataType specifier!\n");
	for(i=0; i<dteltcount; i++)
	  print_scew_element(scew_element_by_index(subelt, i), stderr);
	exit(-1);
      }
    }
    else
      fatal_error("Unrecognized GeolocationField/DataField element:", subelt_name);
  }
  if(dimlist != NULL) free(dimlist);
}

void report_type_mismatch (char const *wanted, char const *saw)
{
  print_name(stderr);
  fprintf(stderr, "\n\texpected type %s, saw type %s\n", wanted, saw);
  bump_errors();
}

void report_range_mismatch(char const *type, unsigned int indexp, char const *value, scew_element const *elt)
{
  print_name(stderr);
  fprintf(stderr, "\n\t%s value [%d] (%s) is out of range(s)\n\t",
	  type, indexp, value);
  print_scew_element(elt, stderr);
  fprintf(stderr, "\n");
  bump_errors();
}

void report_string_mismatch(char const *value, scew_element const *elt) {
  print_name(stderr);
  fprintf(stderr, "\n\tstring \"%s\"\n\tdoesn't match ", value);
  print_scew_element(elt, stderr);
  fprintf(stderr, "\n");
  bump_errors();
}

void report_size_mismatch (char const *type, unsigned long wanted, size_t saw)
{
  print_name(stderr);
  fprintf(stderr, "\n\texpected %s size %ld, saw size %ld\n",
	  type, wanted, (long)saw);
  bump_errors();
}
