/* include the standard headers for this project */
#include "headers.h"

/* the usual predeclaration of functions */

void grind_attributes(scew_element const *elt, hid_t handle, he5_object htype, char const *field);
void validate_attribute (hid_t handle, he5_object htype, char const *field, H5T_class_t type, hsize_t count, size_t size, H5T_sign_t sign, scew_element const *att);
void validate_int_values(void *ival, hsize_t count, size_t tsize, H5T_sign_t signp, scew_element const *elt);
void validate_float_values(void *fval, hsize_t count, size_t tsize, scew_element const *elt);
void validate_string_attribute(char *cval, hsize_t count, scew_element const *att);
void grind_swaths(scew_element const *elt, hid_t handle, char *swnames, int numswath);
void validate_swath(hid_t swhandle, char const *swname, scew_element const *elt);
void validate_dimension(scew_element const *dimelt,
			long numdims, char *dimnames, hsize_t *dimsizes,
			int mandatoryp);
void validate_field(hid_t handle, char const *fieldname, he5_object hetype, scew_element const *glfelt);

void report_type_mismatch (char const *wanted, char const *saw);
void report_range_mismatch(char const *type, int index, char const *value, scew_element const *elt);
void report_string_mismatch(char const *value, scew_element const *elt);
void report_size_mismatch (char const *type, long wanted, size_t saw);

/* finally, the main routine */

/* validate-hdfeos5 -x xml-file -h hdf-eos-file */
/* General philosophy of this program:
   The XML file describes elements in the HDF-EOS file which are examined
   and validated.  If there are extra things in the HDF-EOS file, they
   are not checked, validated, or complained about in any way.
*/
int main(int argc, char **argv)
{
  char *xmlfile = NULL;
  char *hdffile = NULL;
  int i;
  scew_parser *xmlparser;
  hid_t hdfhandle;
  scew_element const *xmlroot;

  for(i=1; i<argc; i++) {
    char *arg = argv[i];
    if(!strcmp(arg, "-x")) xmlfile = argv[++i];
    if(!strcmp(arg, "-h")) hdffile = argv[++i];
  }

  if(xmlfile == NULL || hdffile == NULL) {
    fprintf(stderr, "USAGE: validate-hdfeos5 [-x xml-file] [-h hdf-eos5-file]\n");
    exit(-1);
  }

  xmlparser = scew_parser_create();
  if(!scew_load_file(xmlparser, xmlfile)) {
    XML_Parser exp = scew_parser_expat(xmlparser);

    fprintf(stderr, "FATAL ERROR: %s is unreadable or bogus XML: %s, line %d column %d\n",
	    xmlfile, XML_ErrorString(XML_GetErrorCode(exp)), XML_GetCurrentLineNumber(exp), XML_GetCurrentColumnNumber(exp));
    exit(-1);
  }

  SS(hdfhandle = HE5_SWopen(hdffile, H5F_ACC_RDONLY));

  /* Done with argument processing, XML file reading, HDF-EOS5 file opening */
  /* On with the show! */
  
  /* Spit if the top-level element isn't <HDF-EOS-Validation-Definition> */
  xmlroot = scew_get_root(xmlparser);
  check_element_name(xmlroot, "HDF-EOS-Validation-Definition", 1);

  /* We passed that superficial error check, now to inhale the top-level */
  for(i=0; i< scew_get_element_count(xmlroot); i++) {
    scew_element const *elt = scew_get_element(xmlroot,i);
    char const *elt_name = scew_get_element_name(elt);

    TRACE((stderr, "Processing element named %s\n", elt_name));

    if(!strcmp(elt_name, "Attributes")) {
      push_name("FILE", hdffile);
      grind_attributes(elt, hdfhandle, he5_file, NULL);
      pop_name();
    }

    else if(!strcmp(elt_name, "Swaths")) {
      char *swathnames;
      int numswaths;
      long swathnamesize;

      SS(numswaths = HE5_SWinqswath(hdffile, NULL, &swathnamesize));
      if(scew_get_element_count(elt) > 0 && numswaths == 0)
	fatal_error("No swaths in file:", hdffile);

      swathnames = newstring(swathnamesize);
      SS(numswaths = HE5_SWinqswath(hdffile, swathnames, &swathnamesize));

      grind_swaths(elt, hdfhandle, swathnames, numswaths);

      free(swathnames);
    }
    else {
      fatal_error("Unexpected element inside <HDF-EOS-Validation-Definition>", elt_name);
    }
  }
  return print_final_message();
}


void grind_attributes(scew_element const *elt, hid_t handle, he5_object htype, char const *field)
{
  scew_element const *mandatory = scew_get_element_by_name(elt, "Mandatory");
  scew_element const *optional = scew_get_element_by_name(elt, "Optional");
  int i;
  char *attnamebuf;
  int numatts;
  long bufsize;

  /* fetch the global attribute names */
  SS(attribute_names(handle, htype, field, NULL, &bufsize));
  attnamebuf = newstring(bufsize);
  SS(numatts = attribute_names(handle, htype, field, attnamebuf, &bufsize));

  TRACE((stderr, "Grinding attributes\n"));

  /* mash the mandatory attributes */
  if(mandatory != NULL) {
    for(i=0; i<scew_get_element_count(mandatory); i++) {
      scew_element const *att = scew_get_element(mandatory, i);

      if(!strcmp(scew_get_element_name(att), "Attribute")) {
	scew_attribute const *nameatt = scew_get_attribute_by_name(att, "name");

	if(nameatt == NULL)
	  fatal_error("Attribute with no name:", "");
	else {
	  char const *attname = scew_get_attribute_value(nameatt);
	  H5T_class_t type;
	  hsize_t count;
	  size_t size;
	  H5T_sign_t sign;


	  if(name_in_list(attname, attnamebuf, numatts) < 0) {
	    print_name(stderr);
	    fprintf(stderr, "\n\tMissing mandatory attribute: %s\n",
		    attname);
	    bump_errors();
	  }
	  else {
	    push_name("ATTRIBUTE", attname);
	    SS(attribute_info(handle, htype, field, attname,
			      &type, &count, &size, &sign));
	    validate_attribute(handle, htype, field,
			       type, count, size, sign, att);
	    pop_name();
	  }
	}
      }
      /* should we report non <Attribute ...> elements? */
    }
  }

  /* then the optional attributes */
  if(optional != NULL) {
    for(i=0; i<scew_get_element_count(optional); i++) {
      scew_element const *att = scew_get_element(optional, i);

      if(!strcmp(scew_get_element_name(att), "Attribute")) {
	scew_attribute const *nameatt = scew_get_attribute_by_name(att, "name");

	if(nameatt == NULL)
	  fatal_error("Attribute with no name:", "");
	else {
	  char const *attname = scew_get_attribute_value(nameatt);
	  H5T_class_t type;
	  hsize_t count;
	  size_t size;
	  H5T_sign_t sign;


	  if(name_in_list(attname, attnamebuf, numatts) >= 0) {
	    push_name("OPTIONAL-ATTRIBUTE", attname);
	    SS(attribute_info(handle, htype, field, attname,
			      &type, &count, &size, &sign));
	    validate_attribute(handle, htype, field,
			       type, count, size, sign, att);
	    pop_name();
	  }
	}
      }
      /* should we report non <Attribute ...> elements? */
    }
  }

  /* clean up after ourselves */
  free(attnamebuf);
}

/* Yet another *&^%$#@! function mapping small integers safely to
   strings.  Oh, for a language with REAL enumerated types ...
*/
char const *H5T_class_name(H5T_class_t type)
{
  switch(type) {
  case H5T_STRING:  return "H5T_STRING";
  case H5T_INTEGER: return "H5T_INTEGER";
  case H5T_FLOAT:   return "H5T_FLOAT";
  default:          return "Unknown HDF data type";
  }
}

void validate_attribute (hid_t handle, he5_object htype, char const *field,
			 H5T_class_t ntype, hsize_t count,  size_t size, H5T_sign_t signp, scew_element const *att)
{
  /* the list of value attributes */
  char *valname[] = {"StringValue", "IntValue", "FloatValue",
		     "SameDataTypeAsField", NULL};
  scew_element const *valspec;
  int valindex;

  TRACE((stderr, "Validating %s %s\n", current_type(), current_name()));
  for(valindex=0; valname[valindex] != NULL; valindex++)
    if((valspec = scew_get_element_by_name(att, valname[valindex])) != NULL)
      break;
  if(valspec != NULL) {

    void *valbuf = (void *)malloc(count * size + 1);

    /* the '+1' above is to insure that string attributes are properly
       C-style terminated, as is this next statement:
    */
    ((char *)valbuf)[count * size] = 0;

    
    SS(read_attribute(handle, htype, field, current_name(), valbuf));

    TRACE((stderr, "Looking for %s\n", valname[valindex]));
    switch(valindex) {

    case 0: /* StringValue */
      if(ntype == H5T_STRING) {
	validate_string_attribute((char *)valbuf, count, valspec);
      }
      else report_type_mismatch(valname[valindex], "string");
      break;

    case 1: /* IntValue */
      if(ntype == H5T_INTEGER) {
	validate_int_values(valbuf, count, size, signp, valspec);
      }
      else report_type_mismatch(valname[valindex], "integer");
      break;

    case 2: /* FloatValue */
      if(ntype == H5T_FLOAT) {
	validate_float_values(valbuf, count, size, valspec);
      }
      else report_type_mismatch(valname[valindex], "float");
      break;

    case 3: /* SameDataTypeAsField */
      if (htype == he5_geofield || htype == he5_datafield) {
	int numdims;
	hid_t fieldntype;
	H5T_class_t fieldclass;

	/* grumble, grumble, non-const-using HDF-EOS developers ... */
	SS(HE5_SWfieldinfo(handle, (char *)field, &numdims, NULL, &fieldntype, NULL, NULL));
	fieldclass = type_info(read_in_type(fieldntype), NULL, NULL);

	if(ntype != fieldclass) {
	  char msgbuf[100];
	  sprintf(msgbuf, "SameDataTypeAsField(%s)", H5T_class_name(ntype));
	  report_type_mismatch(msgbuf, H5T_class_name(fieldclass));
	}
      }
      else {
	print_name(stderr);
	fatal_error("\nSameTypeAsDataField on a non-field", "");
      }
      break;
    }
    free(valbuf);
  }
  else {
    print_scew_element(att, stderr);
    fatal_error("\n\tNo Value specifier",
		"(StringValue, IntValue, FloatValue, SameDataTypeAsField)");
  }
}

/* <IntValue><IntRange min="1" max="12"/><Int value="42"/></IntValue> */

void validate_int_values(void *ival, hsize_t count, size_t tsize, H5T_sign_t signp, scew_element const *elt)
{
  struct intpair {
    long min;
    long max;
  };
  struct intpair ipbuf[20];
  struct intpair *ipvec;
  int ipcount = scew_get_element_count(elt);
  int i;
  char valstring[25];
  char const * sizestring = get_attribute_string(elt, "size");

  if(sizestring != NULL) {
    int bogus;
    long xsize = careful_atol(sizestring, H5T_SGN_NONE, &bogus);
    if(!bogus) {
      if(xsize != tsize)
	report_size_mismatch("integer", xsize, tsize);
    }
    else fatal_error("Not an integer size", sizestring);
  }

  if(ipcount > (sizeof(ipbuf)/sizeof(struct intpair)))
    ipvec = calloc(ipcount, sizeof(struct intpair));
  else
    ipvec = ipbuf;

  for(i=0; i<ipcount; i++) {
    scew_element const *subelt = scew_get_element(elt, i);
    char const *subelt_name = scew_get_element_name(subelt);

    if(!strcmp(subelt_name, "IntRange")) {
      char const *numstring = get_attribute_string(subelt, "min");
      int bogus;

      if(numstring != NULL) {
	ipvec[i].min = careful_atol(numstring, signp, &bogus);
	if(bogus)
	  fatal_error("Not a valid integer", numstring);
      }
      else ipvec[i].min = (signp ? LONG_MIN : 0);

      numstring = get_attribute_string(subelt, "max");
      if(numstring != NULL) {
	ipvec[i].max = careful_atol(numstring, signp, &bogus);
	if(bogus)
	  fatal_error("Not a valid integer", numstring);
      }
      else ipvec[i].max = (signp ? LONG_MAX : ULONG_MAX);
    }
    else if (!strcmp(subelt_name, "Int")) {
      char const *numstring = get_attribute_string(subelt, "value");
      if(numstring != NULL) {
	int bogus;

	ipvec[i].min = ipvec[i].max = careful_atol(numstring, signp, &bogus);
	if(bogus)
	  fatal_error("Not a valid integer", numstring);
      }
      else
	fatal_error("No value attribute for Int", "");
    }
    else
      fatal_error("Bogus IntValue spec", subelt_name);
  }

  if(ipcount) {
    if (signp) {
      for(i=0; i<count; i++) {
	long val;
	int j;
	int valueok = 0;
	char *reason = NULL;
	long loser;

	switch(tsize) {
	case 1: val = ((char *)ival)[i]; break;
	case 2: val = ((short *)ival)[i]; break;
	case 4: val = ((int *)ival)[i]; break;
	case 8: val = ((long *)ival)[i]; break;
	default:
	  sprintf(valstring, "%d", tsize);
	  fatal_error("Unknown int size:", valstring);
	}
    
	for(j=0; j<ipcount; j++) {
	  if(val >= ipvec[j].min && val <= ipvec[j].max) {
	    valueok = 1;
	    break;
	  }
	}

	if(!valueok) {
	  sprintf(valstring, "%ld", val);
	  report_range_mismatch("int", i, valstring, elt);
	}
      }
    }
    else {
      for(i=0; i<count; i++) {
	unsigned long val;
	int j;
	int valueok = 0;

	switch(tsize) {
	case 1: val = ((unsigned char *)ival)[i]; break;
	case 2: val = ((unsigned short *)ival)[i]; break;
	case 4: val = ((unsigned int *)ival)[i]; break;
	case 8: val = ((unsigned long *)ival)[i]; break;
	default:
	  sprintf(valstring, "%d", tsize);
	  fatal_error("Unknown int size:", valstring);
	}
	for(j=0; j<ipcount; j++) {
	  if(val >= (unsigned long)ipvec[j].min
	     && val <= (unsigned long)ipvec[j].max) {
	    valueok = 1;
	    break;
	  }
	}

	if(!valueok) {
	  sprintf(valstring, "%uld", val);
	  report_range_mismatch("unsigned int", i, valstring, elt);
	}
      }
    }
    if(ipvec != ipbuf) free(ipvec);
  }
}
/* <FloatValue min="1" max="12"\>
   Omitting min or max means no minimum/maximum value
*/

void validate_float_values(void *fval, hsize_t count, size_t tsize, scew_element const *elt)
{
  struct floatpair {
    double min;
    double max;
  };
  struct floatpair fpbuf[20];
  struct floatpair *fpvec;
  int fpcount = scew_get_element_count(elt);
  char valstring[25];
  char *typename = (tsize == 4 ? "float" : "double");
  int i;
  char const * sizestring = get_attribute_string(elt, "size");

  if(sizestring != NULL) {
    int bogus;
    long xsize = careful_atol(sizestring, H5T_SGN_NONE, &bogus);
    if(!bogus) {
      if(xsize != tsize)
	report_size_mismatch("float", xsize, tsize);
    }
    else fatal_error("Not a float size", sizestring);
  }

  if(fpcount > (sizeof(fpbuf)/sizeof(struct floatpair)))
    fpvec = calloc(fpcount, sizeof(struct floatpair));
  else
    fpvec = fpbuf;

  for(i=0; i<fpcount; i++) {
    scew_element const *subelt = scew_get_element(elt, i);
    char const *subelt_name = scew_get_element_name(subelt);

    if(!strcmp(subelt_name, "FloatRange")) {
      char const *numstring = get_attribute_string(subelt, "min");
      char *bogus = NULL;

      if(numstring != NULL) {
	fpvec[i].min = strtod(numstring, &bogus);
	if(bogus != numstring + strlen(numstring) || (errno == ERANGE))
	  fatal_error("Not a valid float", numstring);
      }
      else fpvec[i].min = -HUGE_VAL;

      numstring = get_attribute_string(subelt, "max");
      if(numstring != NULL) {
	fpvec[i].max = strtod(numstring, &bogus);
	if(bogus != numstring + strlen(numstring) || (errno == ERANGE))
	  fatal_error("Not a valid float", numstring);
      }
      else fpvec[i].max = HUGE_VAL;
    }
    else if (!strcmp(subelt_name, "Float")) {
      char const *numstring = get_attribute_string(subelt, "value");
      if(numstring != NULL) {
	char *bogus = NULL;

	fpvec[i].min = fpvec[i].max = strtod(numstring, &bogus);
	if(bogus != numstring + strlen(numstring) || (errno == ERANGE))
	  fatal_error("Not a valid float", numstring);
      }
      else
	fatal_error("No value attribute for Float", "");
    }
    else
      fatal_error("Bogus FloatValue spec", subelt_name);
  }

  if(fpcount) {
    for(i=0; i<count; i++) {
      int j;
      double val;
      int valueok = 0;

      switch(tsize) {
      case 4: val = ((float *)fval)[i]; break;
      case 8: val = ((double *)fval)[i]; break;
      default:
	sprintf(valstring, "%d", tsize);
	fatal_error("Unknown float size:", valstring);
      }
      for(j=0; j<fpcount; j++) {
	if(val >= fpvec[j].min && val <= fpvec[j].max) {
	  valueok = 1;
	  break;
	}
      }
      if(!valueok) {
	sprintf(valstring, "%f", val);
	report_range_mismatch(typename, i, valstring, elt);
      }
    }
    if(fpvec != fpbuf) free(fpvec);
  }
}

/* <StringValue>
     <OneOf> <s>first</s> <s>second</s> </OneOf>
     <Matches>regular-expression</Matches>
   </StringValue>
   Regular expressions follow POSIX (see man regcomp(3) on any Linux box)
*/
void validate_string_attribute(char *cval, hsize_t count, scew_element const *svelt)
{

  TRACE((stderr, "Validating string: \"%s\" against ", cval));
  TRACEWRAP((print_scew_element(svelt, stderr)));
  TRACE((stderr, "\n"));

  switch (scew_get_element_count(svelt)) {
  case 0:
    /* as a convenience, you can say <StringValue value="foo"/> */
    /* or <StringValue matches="regular-expression"/>           */
    {
      char const *inlineval = get_attribute_string(svelt, "value");
      char const *matchval = get_attribute_string(svelt, "matches");

      if(inlineval != NULL) {
	if(strcmp(cval, inlineval))
	  report_string_mismatch(cval, svelt);
      }
      else if(matchval != NULL) {
	if(!regex_matches(cval, matchval))
	    report_string_mismatch(cval, svelt);
      }
      else {
	TRACE((stderr, "empty StringValue matches anything\n"));
      }
    }
    break;
  case 1:
    {
      scew_element const *subelt = scew_get_element(svelt, 0);
      char const *subeltname = scew_get_element_name(subelt);

      if(!strcmp(subeltname, "OneOf")) {
	int i;
	int foundit = 0;

	for(i=0; i<scew_get_element_count(subelt); i++) {
	  if(!strcmp(cval, scew_get_element_contents(scew_get_element(subelt, i)))) {
	    foundit = 1;
	    TRACE((stderr, "Found OneOf %s\n", cval));
	    break;
	  }
	}
	if(!foundit)
	  report_string_mismatch(cval, subelt);
      }
      else
	fatal_error("Unrecognized StringValue type:", subeltname);
      break;
    }
  default:
    fatal_error("More than one StringValue subtype:", scew_get_element_contents(svelt));
    break;
  }
}

void grind_swaths(scew_element const *elt, hid_t filehandle, char *swnames, int numswaths)
{
  int i;

  TRACE((stderr, "In grind_swaths\n"));

  for(i=0; i<scew_get_element_count(elt); i++) {
    scew_element const *subelt = scew_get_element(elt, i);
    char const *subeltname = scew_get_element_name(subelt);

    if(!strcmp(subeltname, "Every-Swath")) {
      char *swathname = swnames;
      int s;

      for(s=0; s<numswaths; s++) {
	hid_t swhandle;

	end_name(swathname);
	SS(swhandle = HE5_SWattach(filehandle, swathname));
	validate_swath(swhandle, swathname, subelt);
	HE5_SWdetach(swhandle);
	swathname = next_name(swathname);
      }
    }
    else if(!strcmp(subeltname, "Swath-Named")) {
      scew_attribute const *nameatt = scew_get_attribute_by_name(subelt, "name");

      if(nameatt == NULL)
	fatal_error("Named swath with no name:", "");
      else {
	char const *swname = scew_get_attribute_value(nameatt);

	if(name_in_list(swname, swnames, numswaths) < 0) {
	  print_name(stderr);
	  fprintf(stderr, "\n\tMissing mandatory swath: %s\n", swname);
	  bump_errors();
	}
	else {
	  hid_t swhandle;
	  SS(swhandle = HE5_SWattach(filehandle, swname));
	  validate_swath(swhandle, swname, subelt);
	  HE5_SWdetach(swhandle);
	}
      }
    }
    else
      fatal_error("Unrecognized Swath specifier:", subeltname);
  }	  
}

void validate_swath(hid_t swhandle, char const *swname, scew_element const *elt)
{
  int i;

  TRACE((stderr, "Validating swath %s\n", swname));
  push_name("SWATH", swname);
  for(i=0; i < scew_get_element_count(elt); i++) {    
    scew_element const *subelt = scew_get_element(elt,i);
    char const *subelt_name = scew_get_element_name(subelt);

    if(!strcmp(subelt_name, "Attributes"))
      grind_attributes(subelt, swhandle, he5_swath, NULL);

    else if(!strcmp(subelt_name, "Swath-Dimensions")) {
      long numdims;
      long bufsize;
      hsize_t *dimsizes;
      char *dimnames;
      int j;
      

      /* Read in the dimension names and sizes */
      SS(numdims = HE5_SWnentries(swhandle, HE5_HDFE_NENTDIM, &bufsize));
      dimnames = newstring(bufsize + 1);
      dimsizes = (hsize_t *) malloc(sizeof(hsize_t) * numdims);
      SS(numdims = HE5_SWinqdims(swhandle, dimnames, dimsizes));

      /* and validate them against the Dimensions entries */
      for(j=0; j< scew_get_element_count(subelt); j++) {
	scew_element const *manop = scew_get_element(subelt, j);
	char const *manopname = scew_get_element_name(manop);

	if(!strcmp(manopname, "Mandatory-Dimensions")) {
	  int k;

	  for(k=0; k< scew_get_element_count(manop); k++) {
	    validate_dimension(scew_get_element(manop, k),
			       numdims, dimnames, dimsizes,
			       1);
	  }
	}
	else if(!strcmp(manopname, "Optional-Dimensions")) {
	  int k;

	  for(k=0; k< scew_get_element_count(manop); k++) {
	    validate_dimension(scew_get_element(manop, k),
			       numdims, dimnames, dimsizes,
			       0);
	  }
	}
	else
	  fatal_error("Not Mandatory or Optional element:", manopname);
      }
      free(dimsizes);
      free(dimnames);
    }
    else if(!strcmp(subelt_name, "GeolocationFields")) {
      char *glfnames;
      int numglf;
      long glfnamesize;
      int j;

      SS(numglf = HE5_SWnentries(swhandle, HE5_HDFE_NENTGFLD, &glfnamesize));
      if(numglf > 0) {
	glfnames = newstring(glfnamesize + 1);
	SS(numglf = HE5_SWinqgeofields(swhandle, glfnames, NULL, NULL));
      }

      for(j=0; j<scew_get_element_count(subelt); j++) {
	scew_element const *glelt = scew_get_element(subelt, j);
	char const *gleltname = scew_get_element_name(glelt);

	if(!strcmp(gleltname, "Every-GeolocationField")) {
	  char *glfname = glfnames;
	  int k;

	  for(k=0; k<numglf; k++) {
	    end_name(glfname);
	    push_name("GEOLOCATION_FIELD", glfname);
	    validate_field(swhandle, glfname, he5_geofield, glelt);
	    pop_name();
	    glfname = next_name(glfname);
	  }
	}
	else if (!strcmp(gleltname, "GeolocationField-Named")) {
	  char const *glfname = get_attribute_string(glelt, "name");

	  if(glfname != NULL) {
	    if(name_in_list(glfname, glfnames, numglf) >=0) {
	      push_name("GEOLOCATION_FIELD", glfname);
	      validate_field(swhandle, glfname, he5_geofield, glelt);
	      pop_name();
	    }
	    else {
	      print_name(stderr);
	      fprintf(stderr, "\n\tMissing mandatory geolocation field: %s\n",
		      glfname);
	      bump_errors();
	    }
	  }
	  else
	    fatal_error("Named element with no name", "");
	}
	else
	  fatal_error("Unrecognized GeolocationField element", gleltname);
      }
    }
    else if(!strcmp(subelt_name, "DataFields")) {
      char *dfnames;
      int numdf;
      long dfnamesize;
      int j;

      SS(numdf = HE5_SWnentries(swhandle, HE5_HDFE_NENTDFLD, &dfnamesize));
      if(numdf > 0) {
	dfnames = newstring(dfnamesize + 1);
	SS(numdf = HE5_SWinqdatafields(swhandle, dfnames, NULL, NULL));
      }

      for(j=0; j<scew_get_element_count(subelt); j++) {
	scew_element const *dfelt = scew_get_element(subelt, j);
	char const *dfeltname = scew_get_element_name(dfelt);

	if(!strcmp(dfeltname, "Every-DataField")) {
	  char *dfname = dfnames;
	  int k;

	  for(k=0; k<numdf; k++) {
	    end_name(dfname);
	    push_name("DATA_FIELD", dfname);
	    validate_field(swhandle, dfname, he5_datafield, dfelt);
	    pop_name();
	    dfname = next_name(dfname);
	  }
	}
	else if (!strcmp(dfeltname, "DataField-Named")) {
	  char const *dfname = get_attribute_string(dfelt, "name");

	  if(dfname != NULL) {
	    if(name_in_list(dfname, dfnames, numdf) >=0) {
	      push_name("DATA_FIELD", dfname);
	      validate_field(swhandle, dfname, he5_datafield, dfelt);
	      pop_name();
	    }
	    else {
	      print_name(stderr);
	      fprintf(stderr, "\n\tMissing mandatory data field: %s\n",
		      dfname);
	      bump_errors();
	    }
	  }
	  else
	    fatal_error("Named element with no name", "");
	}
	else
	  fatal_error("Unrecognized DataField element", dfeltname);
      }
    }
    else
      fatal_error("Unrecognized Swath element", subelt_name);
  }
  pop_name();
}

void validate_dimension(scew_element const *dimelt,
			long numdims, char *dimnames, hsize_t *dimsizes,
			int mandatoryp)
{
  char const *dimeltname = scew_get_element_name(dimelt);
	    
  if(!strcmp(dimeltname, "Dimension")) {
    char const *dimname = get_attribute_string(dimelt, "name");

    if(dimname == NULL) {
      print_scew_element(dimelt, stderr);
      fatal_error("Dimension element with no name", "");
    }
    else {
      int dimindex = name_in_list(dimname, dimnames, numdims);

      if(dimindex >= 0) {
	char const *sizestr = get_attribute_string(dimelt, "size");
	char const *minstr = get_attribute_string(dimelt, "min");
	char const *maxstr = get_attribute_string(dimelt, "max");
	
	if (sizestr != NULL) {
	  hsize_t size = atoi(sizestr);
	  
	  if(dimsizes[dimindex] != size) {
	    print_name(stderr);
	    fprintf(stderr, "\n\tDimension %s size %lld is not == %lld\n",
		    dimname, dimsizes[dimindex], size);
	    bump_errors();
	  }
	}
	if(minstr != NULL) {
	  hsize_t min = atoi(minstr);
	  
	  if(dimsizes[dimindex] < min) {
	    print_name(stderr);
	    fprintf(stderr, "\n\tDimension %s size %lld is < %lld\n",
		    dimname, dimsizes[dimindex], min);
	    bump_errors();
	  }
	}
	if(maxstr != NULL) {
	  hsize_t max = atoi(maxstr);
	  
	  if(dimsizes[dimindex] > max) {
	    print_name(stderr);
	    fprintf(stderr, "\n\tDimension %s size %lld is > %lld\n",
		    dimname, dimsizes[dimindex], max);
	    bump_errors();
	  }
	}
      }
      else if(mandatoryp) {
	print_name(stderr);
	fprintf(stderr, "\n\tMissing mandatory dimension: %s\n", dimname);
	bump_errors();
      }
    }
  }
  else
    fatal_error("Not a Dimension spec element", dimeltname);
	
}

/* C needs nested functions so bad...*/
void fetch_fieldinfo(hid_t swhandle, char const *fieldname,
		     hid_t *ntype, char **dimlist, hsize_t **dims, int *numdims)
{
  long dimlistsize;

  SS(HE5_SWnentries(swhandle, HE5_HDFE_NENTDIM, &dimlistsize));
  *dimlist = newstring(dimlistsize + 1);

  /* It's not *MY* fault that the HDF-EOS developers don't use const
     properly... heck, I don't use it unless forced by another more
     persnickety library like SCEW...
  */
  SS(HE5_SWfieldinfo(swhandle, (char *)fieldname, numdims, NULL, ntype, *dimlist, NULL));
  *dims = malloc(*numdims * sizeof(hsize_t));
  SS(HE5_SWfieldinfo(swhandle, (char *)fieldname, numdims, *dims, ntype, *dimlist, NULL));
}
  
/* junk function to avoid a little repetitious code loading a field */
void *load_field_data (hid_t swhandle, char const *fieldname, size_t tsize, hsize_t *dimsize, int numdims, hsize_t *valcount)
{
  int i;
  void *databuf;

  *valcount = 1;
  for(i=0; i < numdims; i++)
    *valcount *= dimsize[i];
  databuf = (void *)malloc(*valcount * tsize);
  if(databuf == NULL)
    fatal_error("Couldn't allocate memory to read field", fieldname);
  SS(HE5_SWreadfield(swhandle, (char *)fieldname, NULL, NULL, NULL, databuf));
  return databuf;
}

void validate_field(hid_t swhandle, char const *fieldname, he5_object hetype, scew_element const *glfelt)
{
  int i;
  /* These get filled in on the first occurrence of {Dimensions,DataType},
     and the value gets checked later
  */
  char *dimlist = NULL;
  hid_t ntype;
  int numdims;
  hsize_t *dimsize;

  TRACE((stderr, "Validating %s %s\n", current_type(), current_name()));
  for(i=0; i<scew_get_element_count(glfelt); i++) {    
    scew_element const *subelt = scew_get_element(glfelt,i);
    char const *subelt_name = scew_get_element_name(subelt);

    if(!strcmp(subelt_name, "Attributes"))
      grind_attributes(subelt, swhandle, hetype, fieldname);

    else if(!strcmp(subelt_name, "Dimensions")) {
      int i;

      if(dimlist == NULL)
	fetch_fieldinfo(swhandle, fieldname, &ntype, &dimlist, &dimsize, &numdims);
      for(i=0; i<scew_get_element_count(subelt); i++) {
	char const *dname = scew_get_element_contents(scew_get_element(subelt,i));
	if(name_in_list(dname, dimlist, numdims) < 0) {
	  print_name(stderr);
	  fprintf(stderr, "\n\tMissing mandatory dimension %s\n", dname);
	  bump_errors();
	}
      }
    }
    else if(!strcmp(subelt_name, "DataType")) {
      /* There had better be exactly one subelement, which must be
	 one of IntValue/FloatValue/StringValue */
      int dteltcount = scew_get_element_count(subelt);

      switch(dteltcount) {
      case 1:
	{
	  scew_element const *typelt = scew_get_element(subelt, 0);
	  char const *typeltname = scew_get_element_name(typelt);
	  char const *realtype;
	  H5T_class_t h5type;
	  size_t tsize;
	  H5T_sign_t signp;

	  if(dimlist == NULL)
	    fetch_fieldinfo(swhandle, fieldname, &ntype, &dimlist, &dimsize, &numdims);

	  switch(h5type = type_info(read_in_type(ntype), &tsize, &signp)) {
	  case H5T_STRING:  realtype = "string"; break;
	  case H5T_INTEGER: realtype = "integer"; break;
	  case H5T_FLOAT:   realtype = "float"; break;
	  default:
	    {
	      char msgbuf[25];
	      sprintf(msgbuf, "unknown HDF type %d", ntype);
	      report_type_mismatch(typeltname, msgbuf);
	    }
	  /* note - non-local exit to go around the loop again */
	  continue;
	  }

	  /* if there are no subelements of the typelt, we don't need to read
	     in the data and can just check the size here
	  */
	  if(scew_get_element_count(typelt) == 0) {
	    char const *sizestring = get_attribute_string(typelt, "size");
	    if(sizestring != NULL) {
	      int bogus;
	      long wsize = careful_atol(sizestring, H5T_SGN_NONE, &bogus);
	      if(!bogus) {
		if(wsize != tsize)
		  report_size_mismatch(typeltname, wsize, tsize);
	      }
	      else
		fatal_error("Not a DataType size", sizestring);
	    }
	  }
	  else {

	    if(!strcmp(typeltname, "IntValue")) {
	      if(h5type != H5T_INTEGER)
		report_type_mismatch(typeltname, realtype);
	      else {
		hsize_t intcount;
		void *databuf = load_field_data(swhandle, fieldname, tsize, dimsize, numdims, &intcount);
		validate_int_values(databuf, intcount, tsize, signp, typelt);
		free(databuf);
	      }
	    }
	    else if(!strcmp(typeltname, "FloatValue")) {
	      if(h5type != H5T_FLOAT)
		report_type_mismatch(typeltname, realtype);
	      else {
		hsize_t intcount;
		void *databuf = load_field_data(swhandle, fieldname, tsize, dimsize, numdims, &intcount);
		validate_float_values(databuf, intcount, tsize, typelt);
		free(databuf);
	      }
	    }
	    else if(!strcmp(typeltname, "StringValue")) {
	      if(h5type != H5T_STRING)
		report_type_mismatch(typeltname, realtype);
	      /* punting StringValue validation - I don't think this can
		 actually happen, anyway ... */
	    }
	    else
	      fatal_error("not a DataType specifier", typeltname);
	  }
	}
	break;
      case 0:
	fatal_error("DataType with no type specifier", "");
	break;
      default:
	print_name(stderr);
	fprintf(stderr, "\n\tFATAL ERROR: More than one DataType specifier!\n");
	for(i=0; i<dteltcount; i++)
	  print_scew_element(scew_get_element(subelt, i), stderr);
	exit(-1);
      }
    }
    else
      fatal_error("Unrecognized GeolocationField/DataField element:", subelt_name);
  }
  if(dimlist != NULL) free(dimlist);
}

void report_type_mismatch (char const *wanted, char const *saw)
{
  print_name(stderr);
  fprintf(stderr, "\n\texpected type %s, saw type %s\n", wanted, saw);
  bump_errors();
}

void report_range_mismatch(char const *type, int index, char const *value, scew_element const *elt)
{
  print_name(stderr);
  fprintf(stderr, "\n\t%s value [%d] (%s) is out of range(s)\n\t",
	  type, index, value);
  print_scew_element(elt, stderr);
  fprintf(stderr, "\n");
  bump_errors();
}

void report_string_mismatch(char const *value, scew_element const *elt) {
  print_name(stderr);
  fprintf(stderr, "\n\tstring \"%s\"\n\tdoesn't match ", value);
  print_scew_element(elt, stderr);
  fprintf(stderr, "\n");
  bump_errors();
}

void report_size_mismatch (char const *type, long wanted, size_t saw)
{
  print_name(stderr);
  fprintf(stderr, "\n\texpected %s size %ld, saw size %ld\n",
	  type, wanted, saw);
  bump_errors();
}
