/* include the standard headers for this project */
#include "headers.h"

/* random routines supporting common XMLish operations */

int check_element_name(scew_element const *elt, char const *name, int crap_out)
{
  if(strcmp(name, scew_element_name(elt))) {
    if(crap_out) {
      print_name(stderr);
      fprintf(stderr, "FATAL ERROR: expecting %s in XML but saw %s\n",
	      name, scew_element_name(elt));
      exit(-1);
    }
    else return 1;
  }
  return 0;
}

char const *get_attribute_string(scew_element const *elt, char const *name)
{
  scew_attribute const *att = scew_attribute_by_name(elt, name);
  if(att != NULL) {
    return scew_attribute_value(att);
  }
  else return NULL;
}

/* This just prints a "readable" version of the incoming XML.  It doesn't
   go to the trouble of re-escaping characters, so this isn't a full
   unparser.  The rest of the system uses it when something doesn't match,
   as a convenient label for what it was looking for.
*/
void print_scew_element (scew_element const *elt, FILE *out) {
  char const *name = scew_element_name(elt);
  unsigned int i;

  fprintf(out, "<%s", name);
  for(i=0; i< scew_attribute_count(elt); i++) {
    scew_attribute const *att = scew_attribute_by_index(elt, i);

    fprintf(out, " %s=\"%s\"",
	    scew_attribute_name(att), scew_attribute_value(att));
  }
  if(scew_element_count(elt) == 0
     && scew_element_contents(elt) == NULL) {
    fprintf(out, "/>");
  }
  else {
    fprintf(out, ">");
    for(i=0; i < scew_element_count(elt); i++) {
      print_scew_element(scew_element_by_index(elt, i), out);
    }
    if(scew_element_contents(elt) != NULL) {
      fprintf(out, "%s", scew_element_contents(elt));
    }
    fprintf(out, "</%s>", name);
  }   
}

/* support for storing and retrieving arbitrary values (we do integers as
   longs, floats as doubles, and strings) by name.  The interface looks
   like this:

void save_string(char const *name, char const *val);
void save_int(char const *name, long val);
void save_float(char const *name, double val);

char const *get_saved_string(char const *name);
long const *get_saved_int(char const *name);
double const *get_saved_float(char const *name);

All values are returned as pointers, so NULL uniformly means
nothing-by-that-name.

*/

/* Internal generic node functions */
valnode *new_valnode(void);
valnode *get_saved_node(valnode *vlist, char const *name);
valnode *new_saved_node(valnode **vlist, char const *name);

valnode *new_valnode(void)
{
  return (valnode *)calloc(1, sizeof(valnode));
}

valnode *get_saved_node(valnode *v, char const *name)
{

  for( ; v != NULL; v = v->next)
    if(!strcmp(v->name, name))
      return v;
  return NULL;
}

valnode *new_saved_node(valnode **vlist, char const *name)
{
  valnode *result = get_saved_node(*vlist, name);
  
  if(result == NULL) {
    result = new_valnode();
    result->name = name;
    result->next = *vlist;
    *vlist = result;
  }
  return result;
}

/* The interface to the string memory */
valnode *string_vlist = NULL;

void save_string(char const *name, char const *val)
{
  new_saved_node(&string_vlist, name)->v.strv = val;
}

char const *get_saved_string(char const *name)
{
  valnode *snode = get_saved_node(string_vlist, name);
  return (snode == NULL ? NULL : snode->v.strv);
}

/* The interface to the integer memory */
valnode *int_vlist = NULL;

void save_int(char const *name, long val)
{
  new_saved_node(&int_vlist, name)->v.intv = val;
  TRACE((stderr, "SaveInt for %s is %ld\n", name, val));
}

long const *get_saved_int(char const *name)
{
  valnode *snode = get_saved_node(int_vlist, name);
  return (snode == NULL ? NULL : &(snode->v.intv));
}

/* The interface to the float memory */
valnode *float_vlist = NULL;

void save_float(char const *name, double val)
{
  new_saved_node(&float_vlist, name)->v.floatv = val;
  TRACE((stderr, "SaveFloat for %s is %f\n", name, val));
}

double const *get_saved_float(char const *name)
{
  valnode *snode = get_saved_node(int_vlist, name);
  return (snode == NULL ? NULL : &(snode->v.floatv));
}
